!! Copyright (C) 2004-2010 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module eigensolver_m
  use global
  use oct_parser
  use utilities
  use messages
  use io
  use units
  use mesh
  use quantum_numbers
  use potentials
  use wave_equations_integrator
  use wave_equations
  implicit none


                    !---Derived Data Types---!

  type eigensolver_type
    real(R8) :: tol
  end type eigensolver_type

  type ld_e
     real(R8)            :: e
     real(R8)            :: ldd
     integer             :: nnodes
     type(ld_e), pointer :: next
     type(ld_e), pointer :: prev
  end type ld_e


                    !---Public/Private Statements---!

  private
  public :: eigensolver_type, &
            eigensolver_init, &
            eigensolve, &
            eigensolver_end

contains

  subroutine eigensolver_init(eigensolver)
    !-----------------------------------------------------------------------!
    !-----------------------------------------------------------------------!
    type(eigensolver_type), intent(out) :: eigensolver

    call push_sub("eigensolver_init")

    !Read Eingensolver tolerance
    call oct_parse_double('EigenSolverTolerance', 1.0E-5_r8, eigensolver%tol)
    eigensolver%tol = eigensolver%tol*units_in%energy%factor
    if (eigensolver%tol <= M_ZERO) then
      message(1) = "EigenSolverTolerance must be positive."
      call write_fatal(1)
    end if

    message(1) = ""
    message(2) = "Eigensolver Info"
    call write_info(2)
    write(message(1),'(2X,"Method:    ",A)') "Brent's method"
    write(message(2),'(2X,"Tolerance: ",ES10.3E2)') eigensolver%tol
    call write_info(2,20)

    call pop_sub()
  end subroutine eigensolver_init

  subroutine eigensolver_end(eigensolver)
    !-----------------------------------------------------------------------!
    !-----------------------------------------------------------------------!
    type(eigensolver_type), intent(inout) :: eigensolver

    call push_sub("eigensolver_end")

    !There is nothing to end yet

    call pop_sub()
  end subroutine eigensolver_end

  subroutine eigensolve(qn, occ, wave_eq, m, potential, integrator_sp, &
                        integrator_dp, eigensolver, ev, g, f)
    !-----------------------------------------------------------------------!
    ! Find the eigenvalue and, if requested, compute the corresponding      !
    ! wave-functions.                                                       !
    !                                                                       !
    !  qn            - set of quantum numbers                               !
    !  occ           - occupation                                           !
    !  wave_eq       - wave-equation to use                                 !
    !  m             - mesh                                                 !
    !  potential     - potential to use in the wave-equation                !
    !  integrator_sp - single-precision integrator object                   !
    !  integrator_dp - double-precision integrator object                   !
    !  eigensolver   - information about the eigensolver                    !
    !  ev            - eigenvalue                                           !
    !  g             - wavefunction                                         !
    !  f             - wavefunction                                         !
    !-----------------------------------------------------------------------!
    type(qn_type),          intent(in)    :: qn
    real(R8),               intent(in)    :: occ
    integer,                intent(in)    :: wave_eq
    type(eigensolver_type), intent(in)    :: eigensolver
    type(mesh_type),        intent(in)    :: m
    type(potential_type),   intent(in)    :: potential
    type(integrator_type),  intent(inout) :: integrator_sp, integrator_dp
    real(R8),               intent(out)   :: ev
    real(R8),               intent(inout), optional :: g(m%np), f(m%np)

    integer :: ierr
    type(ld_e), dimension(2) :: eigenvalue_bracket

    call push_sub("eigensolve")

    !Bracket the eigenvalue
    ierr = bracket_eigenvalue(qn, occ, wave_eq, eigensolver%tol, potential, &
                              integrator_sp, eigenvalue_bracket)
    if (ierr /= 0) then
      write(message(1),'("Unable to bracket eigenvalue for state: ",A)') &
           trim(qn_label(qn, .true.))
      call write_fatal(1)
    end if

    !Locate the eigenvalue more accuratly
    ierr = find_eigenvalue(qn, wave_eq, eigensolver%tol, potential, &
                           integrator_dp, eigenvalue_bracket, ev)
    if (ierr /= 0) then
      write(message(1),'("Unable to find eigenvalue for state: ",A)') &
           trim(qn_label(qn, .true.))
      call write_fatal(1)
    end if

    !Get wavefunctions
    if (present(f) .and. present(g)) then
      call wavefunctions(qn, ev, wave_eq, m, potential, integrator_dp, g, f)
    end if

    call pop_sub()
  end subroutine eigensolve

  function bracket_eigenvalue(qn, occ, wave_eq, tol, potential, integrator, eigenvalue_bracket)
    !-----------------------------------------------------------------------!
    ! Bracket the eigenvalue.                                               !
    !                                                                       !
    !  qn                 - set of quantum numbers                          !
    !  occ                - occupation                                      !
    !  wave_eq            - wave-equation to use                            !
    !  tol                - tolerance                                       !
    !  potential          - potential to use in the wave-equation           !
    !  integrator         - integrator object                               !
    !  eigenvalue_bracket - information about upper and lower bounds of     !
    !                       interval containing the eigenvalue              !
    !-----------------------------------------------------------------------!
    type(qn_type),         intent(in)    :: qn
    real(R8),              intent(in)    :: occ
    integer,               intent(in)    :: wave_eq
    real(R8),              intent(in)    :: tol
    type(potential_type),  intent(in)    :: potential
    type(integrator_type), intent(inout) :: integrator
    type(ld_e),            intent(out)   :: eigenvalue_bracket(2)
    integer :: bracket_eigenvalue

    logical             :: unbound
    integer             :: correct_nnodes, last_nnodes
    real(R8)            :: h, last_ldd
    type(ld_e), pointer :: ptr, next, prev, new_ptr
 
    call push_sub("bracket_eigenvalue")

    nullify(ptr, next, prev, new_ptr)
    correct_nnodes = qn_number_of_nodes(qn)
    allocate(ptr);  call ld_e_null(ptr)
    allocate(next); call ld_e_null(next)
    ptr%e = e_min(qn, wave_eq, potential)
    next%e = e_max(qn, wave_eq, tol, potential, ptr%e)
    ptr%next => next
    next%prev => ptr
    h = abs(next%e - ptr%e)
    do
      ptr%ldd = wave_equation_ld_diff(qn, ptr%e, wave_eq, potential, integrator, ptr%nnodes)
      if (.not. associated(ptr%next)) exit
      ptr => ptr%next
    end do
    if (ptr%nnodes < correct_nnodes) then
      eigenvalue_bracket(1)%e = M_ZERO
      eigenvalue_bracket(2)%e = M_ZERO
      call end_ptr(ptr)
      bracket_eigenvalue = 0
      call pop_sub()
      return
    end if

    main: do
      !Check if the eigenvalue has been bracketed
      call goto_last(ptr)
      last_nnodes = ptr%nnodes
      last_ldd = ptr%ldd
      prev => ptr%prev
      do
        if (prev%ldd*ptr%ldd < M_ZERO .and. prev%nnodes == correct_nnodes &
           .and. ptr%nnodes == correct_nnodes) then
          eigenvalue_bracket(1) = prev
          eigenvalue_bracket(2) = ptr
          exit main
        end if
        ptr => prev
        if (.not. associated(ptr%prev)) exit
        prev => ptr%prev
      end do

      !Check for unbound states or some kind of problem
      if (occ == M_ZERO) then
        unbound = .false.
        !Cases were one has unbound states:

        call goto_first(ptr)
        if (last_nnodes == correct_nnodes) then
          ! 1.1 
          if (ptr%nnodes == correct_nnodes .and. ptr%ldd*last_ldd > M_ZERO) unbound = .true.
          ! 1.2
          if (ptr%nnodes == correct_nnodes - 1 .and. ptr%ldd*ptr%next%ldd < M_ZERO &
               .and. ptr%ldd*last_ldd < M_ZERO) unbound = .true.
        end if

        ! 2. The last element has less nodes than the correct number of nodes
        if (last_nnodes < correct_nnodes) unbound = .true.

        if (unbound) then
          !We probably have an unbound state...
          eigenvalue_bracket(1)%e = M_ZERO
          eigenvalue_bracket(2)%e = M_ZERO
          exit main
        end if
      end if

      if (h < sqrt(tol)) then !There is something wrong
        bracket_eigenvalue = 1
        if (in_debug_mode) then
          call bracket_eigenvalue_debug()
          call potential_debug(potential)
        end if
        call end_ptr(ptr)
        call pop_sub()
        return
      end if

      !Refine brackets
      h = h*M_HALF
      call goto_first(ptr)
      next => ptr%next
      do
        allocate(new_ptr)
        new_ptr%e = ptr%e + (next%e - ptr%e)*M_TWOTHIRD
        new_ptr%ldd = wave_equation_ld_diff(qn, new_ptr%e, wave_eq, potential, integrator, new_ptr%nnodes)
        new_ptr%prev => ptr
        new_ptr%next => next
        ptr%next => new_ptr
        next%prev => new_ptr
        nullify(new_ptr)
        ptr => next
        if (.not. associated(ptr%next)) exit
        next => ptr%next
      end do

     !Remove brackets with wrong number of nodes
      call goto_last(ptr)
      prev => ptr%prev
      do
        if (.not. associated(prev%prev)) exit
        if (ptr%nnodes /= correct_nnodes .and. prev%nnodes /= correct_nnodes .and. prev%prev%nnodes == correct_nnodes) then
          nullify(prev%next)
          deallocate(ptr)
        end if
        ptr => prev
        prev => ptr%prev
        if (prev%nnodes == correct_nnodes) exit
      end do

      call goto_first(ptr)
      next => ptr%next
      do
        if (ptr%nnodes == correct_nnodes) exit
        if (next%nnodes <= correct_nnodes-1) then
          nullify(next%prev)
          deallocate(ptr)
        end if
        ptr => next
        if (.not. associated(ptr%next)) exit
        next => ptr%next
        if (next%nnodes >= correct_nnodes) exit
      end do

    end do main

    !Deallocate everything
    call end_ptr(ptr)

    bracket_eigenvalue = 0
    call pop_sub()

  contains
    subroutine bracket_eigenvalue_debug()
      integer :: unit, iostat

      call io_assign(unit)
      open(unit, file='debug_info/bracket_eigenvalue', status='unknown', iostat=iostat)
      if (iostat > 0) then
        message(1) = "Unable to open file ''debug_info/bracket_eigenvalue''"
        call write_fatal(1)
      end if

      call goto_first(ptr)
      do
        write(unit,'(F12.5,1X,ES10.3E2,1X,I2)') ptr%e, ptr%ldd, ptr%nnodes
        if (.not. associated(ptr%next)) exit
        ptr => ptr%next
      end do
      close(unit)

    end subroutine bracket_eigenvalue_debug

  end function bracket_eigenvalue

  subroutine ld_e_null(ptr)
    type(ld_e), pointer :: ptr

    ptr%e = M_ZERO
    ptr%ldd = M_ZERO
    ptr%nnodes = 0
    nullify(ptr%next)
    nullify(ptr%prev)

  end subroutine ld_e_null

  subroutine goto_first(ptr)
    type(ld_e), pointer :: ptr

    do
      if (.not. associated(ptr%prev)) exit
      ptr => ptr%prev
    end do

  end subroutine goto_first

  subroutine goto_last(ptr)
    type(ld_e), pointer :: ptr
        
    do
      if (.not. associated(ptr%next)) exit
      ptr => ptr%next
    end do

  end subroutine goto_last

  subroutine end_ptr(ptr)
    type(ld_e), pointer :: ptr
    
    type(ld_e), pointer :: next

    call goto_first(ptr)
    do
      if (.not. associated(ptr%next)) exit
      next => ptr%next
      nullify(next%prev)
      deallocate(ptr)
      ptr => next
    end do
    deallocate(ptr)

  end subroutine end_ptr

  function find_eigenvalue(qn, wave_eq, tol, potential, integrator, eigenvalue_bracket, ev)
    !-----------------------------------------------------------------------!
    ! Use Brents method to find the eigenvalue.                             !
    !                                                                       !
    !  qn                 - set of quantum numbers                          !
    !  wave_eq            - wave-equation to use                            !
    !  tol                - tolerance                                       !
    !  potential          - potential to use in the wave-equation           !
    !  integrator         - integrator object                               !
    !  eigenvalue_bracket - information about upper and lower bounds of     !
    !                       interval containing the eigenvalue              !
    !  ev                 - eigenvalue                                      !
    !-----------------------------------------------------------------------!
    type(qn_type),         intent(in)    :: qn
    integer,               intent(in)    :: wave_eq
    real(R8),              intent(in)    :: tol
    type(potential_type),  intent(in)    :: potential
    type(integrator_type), intent(inout) :: integrator
    type(ld_e),            intent(in)    :: eigenvalue_bracket(2)
    real(R8),              intent(out)   :: ev
    integer :: find_eigenvalue

    integer  :: nnodes_dum
    real(R8) :: a, b, c, fa, fb, fc, tol1, m, d, e, r, s, q, p, eps = epsilon(M_ONE)

    call push_sub("find_eigenvalue")

    a = eigenvalue_bracket(1)%e
    b = eigenvalue_bracket(2)%e
    if (a == M_ZERO .and. b == M_ZERO) then !Unbound state
      ev = M_ZERO
      find_eigenvalue = 0
      call pop_sub()
      return
    end if
    fa = eigenvalue_bracket(1)%ldd
    fb = eigenvalue_bracket(2)%ldd
    c = b
    fc = fb

    do
      if ((fb < M_ZERO .and. fc < M_ZERO) .or. (fb > M_ZERO .and. fc > M_ZERO)) then
        c = a; fc = fa
        d = b - a; e = d
      end if

      if (abs(fc) < abs(fb)) then
        a = b; b = c; c = a
        fa = fb; fb = fc; fc = fa
      end if
      tol1 = M_TWO*eps*abs(b) + M_HALF*tol
      m = (c - b)*M_HALF

      if (abs(m) <= tol .or. fb == M_HALF) exit

      if (abs(e) > tol .and. abs(fa) > abs(fb)) then
        s = fb/fa
        if (a == c) then !Linear interpolation
          p = M_TWO*m*s
          q = M_ONE - s
        else !Inverse quadratic interpolation
          q = fa/fc
          r = fb/fc
          p = s*(M_TWO*m*q*(q - r) - (b - a)*(r - M_ONE))
          q = (q - M_ONE)*(r - M_ONE)*(s - M_ONE)           
        end if
        if (p > M_ZERO) then
          q = -q
        else
          p = abs(p)
        end if
        s = e
        e = d
        if (M_TWO*p < min(M_THREE*m*q - abs(tol1*q), abs(M_HALF*s*q))) then
          d = p/q
        else
          d = m
          e = d
        end if
      end if

      a = b
      fa = fb
      if (abs(d) > tol1) then
        b = b + d
      else
        b = b + sign(tol1, m)
      end if

      fb = wave_equation_ld_diff(qn, b, wave_eq, potential, integrator, nnodes_dum)
    end do

    ev = b
    find_eigenvalue = 0
    call pop_sub()
  end function find_eigenvalue

  function e_min(qn, wave_eq, potential)
    !-----------------------------------------------------------------------!
    ! Returns the lower bound of the interval where to find the eigenvalue. !
    !-----------------------------------------------------------------------!
    type(qn_type),         intent(in) :: qn
    integer,               intent(in) :: wave_eq
    type(potential_type),  intent(in) :: potential
    real(R8) :: e_min

    real(R8) :: z

    call push_sub("e_min")

    z = potential_nuclear_charge(potential)
    if (z == M_ZERO) then !We probably have a pseudopotential
      e_min = potential_min(potential, qn)
    else !All-electron potential
      e_min = 1.1_r8*hydrogen_eigenvalue(qn, wave_eq, z)
    end if

    call pop_sub()
  end function e_min

  function e_max(qn, wave_eq, tol, potential, emin)
    !-----------------------------------------------------------------------!
    ! Returns the uppper bound of the interval where to find the eigenvalue.!
    !-----------------------------------------------------------------------!
    type(qn_type),         intent(in) :: qn
    integer,               intent(in) :: wave_eq
    real(R8),              intent(in) :: tol
    type(potential_type),  intent(in) :: potential
    real(R8),              intent(in) :: emin
    real(R8) :: e_max

    call push_sub("e_max")

    e_max = min(potential_max(potential, qn), -1e-6_r8)

    call pop_sub()
  end function e_max

end module eigensolver_m
