!! Copyright (C) 2004-2009 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module multireference
  use global
  use gsl_interface
  use messages
  use io
  use utilities
  use output
  use multiroot
  use mesh
  use splines
  use quantum_numbers
  use potentials
  use wave_equations_integrator
  use wave_equations
  use troullier_martins
  use multireference_equations
  implicit none



                    !---Public/Private Statements---!

  private
  public :: mrpp_gen

contains

  subroutine mrpp_gen(qn0, qn1, ev0, ev1, wave_eq, rel, tol, m, ae_potential, &
                      rc0, rc1, integrator, ps_v, ps_g0, ps_f0, ps_g1, ps_f1)
    !-----------------------------------------------------------------------!
    ! Generates the pseudo wave-functions and the corresponding             !
    ! pseudo-potential using the Multireference Pseudo-potentials scheme.   !
    !                                                                       !
    !  qn0          - quantum numbers of semi-core state                    !
    !  qn1          - quantum numbers of valence state                      !
    !  ev0          - eigenvalue of semi-core state                         !
    !  ev1          - eigenvalue of valence state                           !
    !  wave_eq      - wave equation to use                                  !
    !  rel          - if true, use the relativistic extension to the scheme !
    !  tol          - tolerance                                             !
    !  m            - mesh                                                  !
    !  ae_potential - all-electron potential                                !
    !  rc0          - semi-core state cutoff radius                         !
    !  rc1          - valence state cutoff radius                           !
    !  integrator   - integrator object                                     !
    !  ps_v         - pseudo-potential on the mesh                          !
    !  ps_g0        - semi-core state all-electron/pseudo wave-function     !
    !  ps_f0        - semi-core state all-electron/pseudo wave-function     !
    !  ps_g1        - valence state all-electron/pseudo wave-function       !
    !  ps_f1        - valence state all-electron/pseudo wave-function       !
    !                                                                       !
    ! On input qn, ps_f and ps_g contain the all-electron quantum-numbers   !
    ! and wavefunctions and on exit the pseudo wavefunctions and quantum    !
    ! numbers.                                                              !
    !-----------------------------------------------------------------------!
    type(qn_type),         intent(inout) :: qn0, qn1
    real(R8),              intent(in)    :: ev0, ev1
    integer,               intent(in)    :: wave_eq
    logical,               intent(in)    :: rel
    real(R8),              intent(in)    :: tol
    type(mesh_type),       intent(in)    :: m
    type(potential_type),  intent(in)    :: ae_potential
    real(R8),              intent(in)    :: rc0, rc1
    type(integrator_type), intent(inout) :: integrator
    real(R8),              intent(out)   :: ps_v(m%np)
    real(R8),              intent(inout) :: ps_g0(m%np), ps_f0(m%np), ps_g1(m%np), ps_f1(m%np)

    integer :: i
    real(R8) :: g0rc0, gp0rc0, g1rc1, norm0
    real(R8) :: rhs(9), c(9)
    real(R8), allocatable :: ae_g0(:), ae_gp0(:), ae_f0(:)
    real(R8), allocatable :: ae_g1(:), ae_gp1(:), ae_f1(:)
    type(mesh_type) :: m0, m1
    type(spline_type) :: spl

    call push_sub("mrpp_gen")

    !
    call mesh_null(m0)
    call mesh_null(m1)
    call spline_null(spl)

    !Write core radii
    write(message(1),'(4x,"Core radius:",1x,f7.3,5x,f7.3)') rc0, rc1
    call write_info(1,20)
    call write_info(1,unit=out_pp)

    !Copy all-electron wavefunctions
    allocate(ae_g0(m%np), ae_gp0(m%np), ae_f0(m%np))
    allocate(ae_g1(m%np), ae_gp1(m%np), ae_f1(m%np))
    ae_g0 = ps_g0; ae_f0 = ps_f0
    ae_g1 = ps_g1; ae_f1 = ps_f1
    select case (wave_eq)
    case (SCHRODINGER, SCALAR_REL)
      ae_gp0 = ae_f0
      ae_f0  = M_ZERO
      ae_gp1 = ae_f1
      ae_f1  = M_ZERO
    case (DIRAC)
      ae_gp0 = spline_deriv(m%np, m%r, ae_g0, 3)
      ae_gp1 = spline_deriv(m%np, m%r, ae_g1, 3)
    end select

    !Semi-core wavefunctions and norm at rc0
    call spline_init(spl, m%np, m%r, ae_g0, 3)
    g0rc0  = rc0*spline_eval(spl, rc0)
    call spline_end(spl)
    call spline_init(spl, m%np, m%r, ae_gp0, 3)
    gp0rc0 = rc0*spline_eval(spl, rc0) + g0rc0/rc0
    call spline_end(spl)
    if (g0rc0 < M_ZERO) then
      ae_g0  = -ae_g0
      ae_gp0 = -ae_gp0
      ae_f0  = -ae_f0
      g0rc0  = -g0rc0
      gp0rc0 = -gp0rc0
    end if
    if (rel) then
      norm0 = log(spline_integ(m%np, m%r, (ae_g0**2 + ae_f0**2)*m%r**2, M_ZERO, rc0, 3))
    else
      select case (wave_eq)
      case (SCHRODINGER, SCALAR_REL)
        norm0 = log(spline_integ(m%np, m%r, (m%r*ae_g0)**2, M_ZERO, rc0, 3))
      case (DIRAC)
        norm0 = log(spline_integ(m%np, m%r, (m%r*ae_g0)**2, M_ZERO, rc0, 3) + &
                    spline_integ(m%np, m%r, (m%r*ae_f0)**2, M_ZERO, m%r(m%np), 3))
      end select
    end if

    !Valence wavefunctions at rc1
    call spline_init(spl, m%np, m%r, ae_g1, 3)
    g1rc1  = rc1*spline_eval(spl, rc1)
    call spline_end(spl)

    !Get right hand side of equations to solve
    call tm_equations_rhs(m, rc0, rel, qn0, ev0, g0rc0, gp0rc0, norm0, ae_potential, rhs(1:7))
    rhs(8) = g1rc1
    rhs(9) = M_ZERO

    !New meshes
    m0%type = m%type
    m1%type = m%type
    m0%np = locate(m%r, rc0, 0) + 1
    m1%np = locate(m%r, rc1, 0) + 1
    allocate(m0%r(m0%np), m1%r(m1%np))
    m0%r(1:m0%np-1) = m%r(1:m0%np-1)
    m0%r(m0%np) = rc0
    m1%r(1:m1%np-1) = m%r(1:m1%np-1)
    m1%r(m1%np) = rc1

    !Set the quantum numbers to their correct value
    qn0%n = qn0%l + 1
    qn1%n = qn1%l + 2

    !Solve the system of equations
    call mrpp_solve_system(m, m0, m1, .false., wave_eq, integrator, ae_potential,&
         rc0, rc1, qn0, qn1, ev0, ev1, rhs, tol, c, ps_g0(1:m0%np), &
         ps_f0(1:m0%np), ps_g1(1:m1%np), ps_f1(1:m1%np), ps_v(1:m0%np))

    !Write info
    write(message(1),'(6x,"Coefficients:")')
    write(message(2),'(8x,"c0  =",1x,es16.9e2,3x,"c2  =",1x,es16.9e2)') c(1), c(2)
    write(message(3),'(8x,"c4  =",1x,es16.9e2,3x,"c6  =",1x,es16.9e2)') c(3), c(4)
    write(message(4),'(8x,"c8  =",1x,es16.9e2,3x,"c10 =",1x,es16.9e2)') c(5), c(6)
    write(message(5),'(8x,"c12 =",1x,es16.9e2,3x,"c14 =",1x,es16.9e2)') c(7), c(8)
    write(message(6),'(8x,"c16 =",1x,es16.9e2)') c(9)
    message(7) = ""
    call write_info(7,20)
    call write_info(7,unit=out_pp)

    !Compute the polynomial, the pseudo-wavefunction and the pseudopotential
    ps_g0(m0%np:m%np) = ae_g0(m0%np:m%np)
    ps_g1(m1%np:m%np) = ae_g1(m1%np:m%np)
    select case (wave_eq)
    case (SCHRODINGER, SCALAR_REL)
      ps_f0(m0%np:m%np) = ae_gp0(m0%np:m%np)
      ps_f1(m1%np:m%np) = ae_gp1(m1%np:m%np)
    case (DIRAC)
      if (rel) then
        ps_f0(m0%np:m%np) = ae_f0(m0%np:m%np)
        ps_f1(m1%np:m%np) = ae_f1(m1%np:m%np)
      else
        ps_f0 = M_ZERO
        ps_f1 = M_ZERO
      end if
    end select
    do i = m0%np, m%np
      ps_v(i) = v(ae_potential, m%r(i), qn0)
    end do

    !Free memory
    deallocate(ae_g0, ae_gp0, ae_f0, ae_g1, ae_gp1, ae_f1)
    call mesh_end(m0)
    call mesh_end(m1)

    call pop_sub()
  end subroutine mrpp_gen

  subroutine mrpp_solve_system(m, m0, m1, rel, wave_eq, integrator, ae_potential,&
                               rc0, rc1, qn0, qn1, ev0, ev1, rhs, tol, c, &
                               ps_g0, ps_f0, ps_g1, ps_f1, ps_v)
    !-----------------------------------------------------------------------!
    ! Solve the MRPP set of non-linear equations.                           !
    !                                                                       !
    !  m            - mesh                                                  !
    !  m0           - truncated mesh for semi-core state                    !
    !  m1           - truncated mesh for valence state                      !
    !  rel          - if true, use the relativistic extension to the scheme !
    !  wave_eq      - wave equation to use                                  !
    !  integrator   - integrator object                                     !
    !  ae_potential - all-electron potential                                !
    !  rc0          - semi-core state cutoff radius                         !
    !  rc1          - valence state cutoff radius                           !
    !  qn0          - quantum numbers of semi-core state                    !
    !  qn1          - quantum numbers of valence state                      !
    !  ev0          - eigenvalue of semi-core state                         !
    !  ev1          - eigenvalue of valence state                           !
    !  rhs          - right-hand side of equations                          !
    !  tol          - tolerance                                             !
    !  c            - coeficients of the polynomial                         !
    !  ps_g0        - semi-core state pseudo wave-function                  !
    !  ps_f0        - semi-core state pseudo wave-function                  !
    !  ps_g1        - valence state pseudo wave-function                    !
    !  ps_f1        - valence state pseudo wave-function                    !
    !  ps_v         - pseudo-potential                                      !
    !-----------------------------------------------------------------------!
    type(mesh_type),       intent(in)    :: m, m0, m1
    logical,               intent(in)    :: rel
    integer,               intent(in)    :: wave_eq
    type(integrator_type), intent(inout) :: integrator
    type(potential_type),  intent(in)    :: ae_potential
    type(qn_type),         intent(in)    :: qn0, qn1
    real(R8),              intent(in)    :: rc0, rc1, ev0, ev1, rhs(9), tol
    real(R8),              intent(out)   :: c(9)
    real(R8),              intent(out)   :: ps_g0(m0%np), ps_f0(m0%np), ps_g1(m1%np), ps_f1(m1%np), ps_v(m0%np)

    real(R8) :: ldd, cc(4)
    type(multiroot_solver_type) :: mr_solver

    call push_sub("mrpp_solve_system")

    !Solve TM system: we will use it as a starting point
    c = M_ZERO
    call tm_solve_system(m0, rc0, rel, qn0, ev0, rhs(1:7), tol, c(1:7), ps_g0, ps_f0, ps_v, .true.)

    !Allocate memory
    call mrpp_equations_init(m, m0, m1, rel, wave_eq, integrator, ae_potential, rc0, rc1, qn0, qn1, ev0, ev1, rhs)

    !Solve system
    cc(1) = c(1)
    cc(2:4) = c(7:9)
    call multiroot_solver_init(4, M_ZERO, M_ZERO, tol, 300, BROYDEN, mr_solver)
    call multiroot_solve_system(mr_solver, cc, mrpp_equations, mrpp_equations_write_info)

    !Get the final coefficients
    c(1) = cc(1)
    c(7:9) = cc(2:4)
    call mrpp_solve_linear_system(c)

    !Compute the final pseudo-wavefunction and the pseudopotential
    call mrpp_ps_wavefunctions(c, ps_g0, ps_f0, ps_v, ldd, ps_g1, ps_f1)

    !Free memory
    call mrpp_equations_end()

    call pop_sub()
  end subroutine mrpp_solve_system

end module multireference
