!! Copyright (C) 2004-2009 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module run_ape
  use global
  use oct_parser
  use utilities
  use messages
  use units
  use output
  use wave_equations_integrator
  use eigensolver_m
  use atom
  implicit none


                    !---Global Variables---!

  !The stack
  integer :: stack(100) = 0
  integer :: n_inst     = 0

  !Modes
  integer, parameter :: AE         = 1, &
                        PP_GEN     = 2, &
                        PP_TEST    = 4, &
                        PP_CONVERT = 8

  !Instructions
  integer, parameter :: M_AE_ATOM_CREATE  = 1,  &
                        M_AE_ATOM_SAVE    = 2,  &
                        M_AE_ATOM_LOAD    = 3,  &
                        M_AE_ATOM_END     = 4,  &
                        M_PS_ATOM_CREATE  = 11, &
                        M_PS_ATOM_SAVE    = 12, &
                        M_PS_ATOM_LOAD    = 13, &
                        M_PS_ATOM_TEST    = 14, &
                        M_PS_ATOM_END     = 15, &
                        M_KB_ATOM_END     = 16, &
                        M_OUTPUT          = 20, &
                        M_EXIT            = 100


                    !---Public/Private Statements---!

  private
  public :: run, AE, PP_GEN, PP_TEST


contains

  subroutine run()
    !-----------------------------------------------------------------------!
    ! This subroutine reads the run mode and puts the corresponding tasks   !
    ! in a stack. It then performs all the tasks that are in the stack by   !
    ! calling the appropriate subroutines.                                  !
    !                                                                       !
    !-----------------------------------------------------------------------!
    integer                :: calcmode, mode_test, inst, message_length
    type(integrator_type)  :: integrator_sp, integrator_dp
    type(eigensolver_type) :: eigensolver
    type(atom_type)        :: ae_atom, ps_atom, kb_atom

    call push_sub("run")

    !Read calculation mode
    call oct_parse_int('CalculationMode', AE + PP_GEN, calcmode)

    !Write what the code is going to do and check the input
    message(1) = "Calculation Type:"
    message_length = 1
    mode_test = calcmode
    if (iand(calcmode, AE) /= 0) then
      message_length = message_length + 1
      message(message_length) = "  Atomic Calculation"
      mode_test = mode_test - iand(AE, calcmode)
    end if
    if (iand(calcmode, PP_GEN) /= 0) then
      message_length = message_length + 1
      message(message_length) = "  PseudoPotential Generation"
      mode_test = mode_test - iand(PP_GEN, calcmode)
    end if
    if (iand(calcmode, PP_TEST) /= 0) then
      message_length = message_length + 1
      message(message_length) = "  PseudoPotential Test"
      mode_test = mode_test - iand(PP_TEST, calcmode)
    end if
    if (iand(calcmode, PP_CONVERT) /= 0) then
      message_length = message_length + 1
      message(message_length) = "  PseudoPotential File Conversion"
      mode_test = mode_test - iand(PP_CONVERT, calcmode)
    end if
    if (mode_test /= 0) then
      write(message(1), '(A,I2,A)') "Input: '", calcmode, "' is not a valid CalculationMode"
      call write_fatal(1)
    else
      call write_info(message_length)
    end if

    !Initialize stack
    !Remember this is a stack: first in last out!
    call push(M_EXIT) !Last thing to do is always to exit

    !Are we going to convert the pseudopotential data?
    if (iand(calcmode, PP_CONVERT) /= 0) then
      call push(M_PS_ATOM_END)
      call push(M_PS_ATOM_LOAD)
    end if

    !Are we going to test the pseudopotentials?
    if (iand(calcmode, PP_TEST) /= 0) then
      call push(M_PS_ATOM_END)
      call push(M_PS_ATOM_TEST)
      call push(M_PS_ATOM_LOAD)
    end if

    !Are we going to do a pseudopotential generation?
    if (iand(calcmode, PP_GEN) /= 0) then
      call push(M_AE_ATOM_END)
      call push(M_PS_ATOM_END)
      call push(M_KB_ATOM_END)
      call push(M_PS_ATOM_SAVE)
      call push(M_PS_ATOM_CREATE )
      call push(M_AE_ATOM_LOAD)
    end if

    !Are we going to do an all-electron calculation?
    if (iand(calcmode, AE) /= 0) then
      call push(M_AE_ATOM_END)
      call push(M_AE_ATOM_SAVE )
      call push(M_AE_ATOM_CREATE)
    end if

    !Initialize things needed for all modes
    call units_init()
    call wave_equations_integrator_init(integrator_sp, integrator_dp)
    call eigensolver_init(eigensolver)

    !Main loop: do everything that is in the stack
    do
      call pop(inst)
      select case (inst)
      case (M_AE_ATOM_CREATE)
        message(1) = ""
        message(2) = ""
        message(3) = str_center("-- All Electron Calculation --", 70)
        call write_info(3)

        call output_ae_init()
        call atom_null(ae_atom)
        call atom_create_ae(ae_atom, integrator_sp, integrator_dp, eigensolver)
        call atom_output(ae_atom, "ae")

      case (M_AE_ATOM_SAVE)
        call atom_save(ae_atom, "ae")

      case (M_AE_ATOM_LOAD)
        call atom_null(ae_atom)
        call atom_load(ae_atom, "ae")

      case (M_AE_ATOM_END)
        call atom_end(ae_atom)

      case (M_PS_ATOM_CREATE)
        message(1) = ""
        message(2) = ""
        message(3) = str_center("-- Pseudopotential Generation --", 70)
        call write_info(3)

        call output_pp_init()
        call output_kb_init()
        call atom_null(ps_atom)
        call atom_null(kb_atom)
        call atom_create_ps(ae_atom, ps_atom, kb_atom, integrator_sp, &
                            integrator_dp, eigensolver)
        call atom_output(ps_atom, "pp")
        call atom_output(kb_atom, "kb")

      case (M_PS_ATOM_SAVE)
        call atom_save(ps_atom, "pp")

      case (M_PS_ATOM_LOAD)
        call atom_null(ps_atom)
        call atom_load(ps_atom, "pp")

      case (M_PS_ATOM_TEST)
        message(1) = ""
        message(2) = ""
        message(3) = str_center("-- Pseudopotential Testing --", 70)
        call write_info(3)

        call output_tests_init()
        call atom_test(ps_atom, integrator_sp, integrator_dp, eigensolver)
      case (M_PS_ATOM_END)
        call atom_end(ps_atom)
      case (M_EXIT)
        exit
      end select
    end do

    !End things needed for all modes
    call wave_equations_integrator_end(integrator_sp, integrator_dp)
    call eigensolver_end(eigensolver)
    call output_end()

    call pop_sub()
  contains

    subroutine push(inst)
      integer, intent(in) :: inst

      if (n_inst == 100) then
        message(1) = "Too many instructions in stack"
        call write_fatal(1)
      end if
      n_inst = n_inst + 1
      stack(n_inst) = inst

    end subroutine push

    subroutine pop(inst)
      integer, intent(out) :: inst

      if (n_inst == 0) then
        message(1) = "Not enough instructions in stack"
        call write_fatal(1)
      end if
      inst = stack(n_inst)
      stack(n_inst) = 0
      n_inst = n_inst - 1

    end subroutine pop

  end subroutine run

end module run_ape
